"""Type conversion utilities used for both requests and responses"""

import re
from datetime import date, datetime
from io import BytesIO
from logging import getLogger
from os import SEEK_END
from pathlib import Path
from typing import IO, Any, Dict, List, Mapping, MutableSequence, Optional, Union
from warnings import catch_warnings, simplefilter

from dateutil.parser import UnknownTimezoneWarning
from dateutil.parser import parse as parse_date
from dateutil.tz import tzlocal
from requests import Session

from pyinaturalist.constants import (
    MAX_FILESIZE,
    AnyFile,
    Coordinates,
    Dimensions,
    HistogramResponse,
    JsonResponse,
    ResponseResult,
)

GENERIC_TIME_FIELDS = ('created_at', 'last_post_at', 'updated_at')
OBSERVATION_TIME_FIELDS = (
    'created_at_details',
    'created_time_zone',
    'observed_on',
    'observed_on_details',
    'observed_on_string',
    'observed_time_zone',
    'time_zone_offset',
)

# Extremely simplified URL regex, just enough to differentiate from local paths
URL_PATTERN = re.compile(r'^https?://.+')

logger = getLogger(__name__)


# Wrapper functions to apply conversions to all results in a response
# --------------------


def convert_all_coordinates(results: List[ResponseResult]) -> List[ResponseResult]:
    """Convert coordinate pairs in response items from strings to floats, if valid

    Args:
        results: Results from API response; expects coordinates in either 'location' key or
            'latitude' and 'longitude' keys
    """
    results = [convert_lat_long_dict(result) for result in results]
    results = [convert_lat_long_list(result) for result in results]
    return results


def convert_all_place_coordinates(response: JsonResponse) -> JsonResponse:
    """Convert locations for both standard and community-contributed places to floats"""
    response['results'] = {
        'standard': convert_all_coordinates(response['results'].get('standard')),
        'community': convert_all_coordinates(response['results'].get('community')),
    }
    return response


def convert_all_timestamps(results: List[ResponseResult]) -> List[ResponseResult]:
    """Replace all date/time info with datetime objects, where possible"""
    results = [convert_generic_timestamps(result) for result in results]
    results = [convert_observation_timestamps(result) for result in results]
    return results


def convert_observation_histogram(response: JsonResponse) -> HistogramResponse:
    """Convert a response containing time series data into a single ``{date: value}`` dict"""
    # The inner result object's key will be the name of the interval requested
    interval = next(iter(response['results'].keys()))
    return convert_histogram(response['results'][interval], interval)


def convert_histogram(result: JsonResponse, interval: str):
    """Convert keys to appropriate type depending on interval"""
    if interval in ['month_of_year', 'week_of_year']:
        return {int(k): v for k, v in result.items()}
    else:
        return {parse_date(k): v for k, v in result.items()}


# Type conversion functions for individual results or values
# --------------------


def convert_isoformat(value: Union[date, datetime, str]) -> str:
    """Convert a date, datetime, or timestamps to a string in ISO 8601 format.
    If it's a datetime and doesn't already have tzinfo, set it to the system's local timezone.

    Raises:
        :py:exc:`dateutil.parser._parser.ParserError` if a date/datetime format is invalid
    """
    if isinstance(value, str):
        value = parse_date(value)
    if isinstance(value, datetime) and not value.tzinfo:
        value = value.replace(tzinfo=tzlocal())
    return value.isoformat()


def convert_lat_long(obj: Union[Dict, List, None, str]) -> Optional[Coordinates]:
    """Convert a coordinate pair as a dict, list, or string into a pair of floats, if valid"""
    if not obj:
        return None
    elif isinstance(obj, str):
        return try_float_pair(*str(obj).split(','))
    elif isinstance(obj, (list, tuple)):
        return try_float_pair(*obj)
    elif isinstance(obj, dict):
        return try_float_pair(obj.get('latitude'), obj.get('longitude'))


def convert_lat_long_dict(result: ResponseResult) -> ResponseResult:
    """Convert a coordinate pair dict within a response to floats, if valid"""
    if 'latitude' in result and 'longitude' in result:
        result['latitude'] = try_float(result['latitude'])
        result['longitude'] = try_float(result['longitude'])
    return result


def convert_lat_long_list(result: ResponseResult):
    """Convert a coordinate pairs in a response item from strings to floats, if valid"""
    # Format inner record if present, e.g. for search results
    if 'record' in result:
        result['record'] = convert_lat_long_list(result['record'])
        return result

    if ',' in (result.get('location') or ''):
        result['location'] = [try_float(coord) for coord in result['location'].split(',')]
    return result


def convert_generic_timestamps(result: ResponseResult) -> ResponseResult:
    """Replace generic created/updated info that's returned by multiple endpoints.
    **Note:** Compared to observation timestamps, these are generally more reliable. These seem to
    be consistently in ISO 8601 format.
    """
    if not result:
        return result

    # Format inner record if present, e.g. for search results
    if 'record' in result:
        result['record'] = convert_generic_timestamps(result['record'])
        return result

    for field in GENERIC_TIME_FIELDS:
        datetime_obj = try_datetime(result.get(field, ''))
        if datetime_obj:
            result[field] = datetime_obj
    return result


def convert_observation_timestamps(result: ResponseResult) -> ResponseResult:
    """Replace observation date/time info with datetime objects"""
    observation = result.copy()

    observed_on = observation.pop('time_observed_at', None)
    if not isinstance(observation.get('observed_on'), datetime) and observed_on:
        observation['observed_on'] = try_datetime(observed_on)

    created_at = observation.get('created_at')
    if not isinstance(created_at, datetime):
        observation['created_at'] = try_datetime(created_at)

    return observation


def safe_split(value: Any, delimiter: str = '|') -> List[str]:
    """Split a pipe-(or other token)-delimited string"""
    return list(ensure_list(value, split_str_list=True, delimiter=delimiter))


# Type conversion functions for files and collections
# -------------------


def ensure_file_obj(value: AnyFile, session: Optional[Session] = None) -> IO:
    """Load data into a file-like object, if it isn't already. Accepts local file paths and URLs."""
    # Load from URL
    if isinstance(value, str) and URL_PATTERN.match(value):
        session = session or Session()
        file_obj: IO = session.get(value).raw  # type: ignore  # HTTPResponse acts as an IO class
    # Load from local file path
    elif isinstance(value, (str, Path)):
        file_path = Path(value).expanduser().resolve()
        logger.info(f'Reading from file: {file_path}')
        file_obj = BytesIO(file_path.read_bytes())
    # Otherwise, assume it's already a file or file-like object
    elif hasattr(value, 'read'):
        file_obj = value
    else:
        file_obj = BytesIO(value)  # type: ignore

    # Verify maximum file size
    file_obj.seek(0, SEEK_END)
    file_size = file_obj.tell()
    file_obj.seek(0)
    if file_size > MAX_FILESIZE:
        raise ValueError(f'File size exceeds maximum allowed ({MAX_FILESIZE} bytes): {file_size}')

    return file_obj


def ensure_list(
    value: Any, split_str_list: bool = False, delimiter: str = ','
) -> MutableSequence[Any]:
    """Convert an object, response, or (optionally) comma-separated string into a list"""
    # Handle the occacsional stray numpy array that got lost and made its way here
    try:
        value = value.tolist()
    except (AttributeError, TypeError):
        pass
    if not value:
        return []

    # Handle response results and comma-separated strings
    elif isinstance(value, dict) and 'results' in value:
        value = value['results']
    elif split_str_list and isinstance(value, str) and delimiter in value:
        return [s.strip() for s in value.split(delimiter)]

    if isinstance(value, MutableSequence):
        return value
    elif isinstance(value, (tuple, set)):
        return list(value)
    else:
        return [value]


# Formatting functions
# --------------------


def format_dimensions(dimensions: Union[Dict[str, int], Dimensions, None]) -> Dimensions:
    """Simplify a 'dimensions' dict into a ``(width, height)`` tuple"""
    if not dimensions:
        return (0, 0)
    if isinstance(dimensions, (tuple, list, set)):
        return (dimensions[0], dimensions[1])
    elif isinstance(dimensions, Mapping):
        return dimensions.get('width', 0), dimensions.get('height', 0)


def format_file_size(n_bytes: int) -> str:
    """Convert a file size in bytes into a human-readable format"""
    filesize = float(n_bytes or 0)

    def _format(unit):
        return f'{int(filesize)} {unit}' if unit == 'bytes' else f'{filesize:.2f} {unit}'

    for unit in ['bytes', 'KB', 'MB', 'GB']:
        if filesize < 1024 or unit == 'GB':
            return _format(unit)
        filesize /= 1024

    return _format(unit)


def format_license(value: str) -> Optional[str]:
    """Format a Creative Commons license code"""
    return str(value).upper().replace('_', '-') if value else None


# 'Safe' conversion functions that return invalid values as None instead of raising an error
# --------------------


def try_datetime(timestamp: Any, **kwargs) -> Optional[datetime]:
    """Parse a date/time string into a datetime, if valid; return ``None`` otherwise"""
    if isinstance(timestamp, datetime):
        return timestamp
    if not timestamp or not str(timestamp).strip():
        return None

    try:
        # Suppress UnknownTimezoneWarning
        with catch_warnings():
            simplefilter('ignore', category=UnknownTimezoneWarning)
            return parse_date(timestamp, **kwargs)
    except (AttributeError, TypeError, ValueError) as e:
        logger.debug(f'Could not parse timestamp: {timestamp}: "{str(e)}"')
        return None


def try_date(timestamp: Any, **kwargs) -> Optional[date]:
    """Parse a date string into a date, if valid; return ``None`` otherwise"""
    dt = try_datetime(timestamp, **kwargs)
    return dt.date() if dt else None


def try_float(value: Any) -> Optional[float]:
    """Convert a value to a float, if valid; return ``None`` otherwise"""
    try:
        return float(value)
    except (TypeError, ValueError):
        return None


def try_float_pair(*values: Any) -> Optional[Coordinates]:
    """Convert a pair of coordinat values to floats, if both are valid; return ``None`` otherwise"""
    if len(values) != 2:
        return None
    try:
        return float(values[0]), float(values[1])
    except (TypeError, ValueError):
        return None


def try_int(value: Any) -> Optional[float]:
    """Convert a value to a int, if valid; return ``None`` otherwise"""
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def try_int_or_float(value: Any) -> Union[int, float, None]:
    """Convert a value to either an int or a float, if valid; return ``None`` otherwise"""
    return try_int(str(value)) or try_float(str(value))



"""Controller classes for :py:class:`.iNatClient`. These contain all the request functions used by
the client, grouped by resource type.
"""

# ruff: noqa: F401
# isort: skip_file
from pyinaturalist.controllers.base_controller import BaseController
from pyinaturalist.controllers.annotation_controller import AnnotationController
from pyinaturalist.controllers.observation_controller import ObservationController
from pyinaturalist.controllers.place_controller import PlaceController
from pyinaturalist.controllers.project_controller import ProjectController
from pyinaturalist.controllers.taxon_controller import TaxonController
from pyinaturalist.controllers.user_controller import UserController
from pyinaturalist.controllers.search_controller import SearchController



from typing import Optional

from pyinaturalist.constants import API_V1, IntOrStr, JsonResponse, MultiInt
from pyinaturalist.converters import convert_all_timestamps
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.paginator import paginate_all
from pyinaturalist.request_params import convert_rank_range
from pyinaturalist.session import get


@document_request_params(docs._taxon_params, docs._taxon_id_params, docs._pagination)
def get_taxa(**params) -> JsonResponse:
    """Search taxa

    .. rubric:: Notes

    * API reference: :v1:`GET /taxa <Taxa/get_taxa>`

    Example:
        >>> response = get_taxa(q='vespi', rank=['genus', 'family'])
        >>> pprint(response)
        [52747] Family: Vespidae (Hornets, Paper Wasps, Potter Wasps, and Allies)
        [92786] Genus: Vespicula
        [646195] Genus: Vespiodes
        ...

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_taxa.json
                :language: JSON

    Returns:
        Response dict containing taxon records
    """
    params = convert_rank_range(params)
    if params.get('page') == 'all':
        taxa = paginate_all(get, f'{API_V1}/taxa', **params)
    else:
        taxa = get(f'{API_V1}/taxa', **params).json()

    taxa['results'] = convert_all_timestamps(taxa['results'])
    return taxa


def get_taxa_by_id(
    taxon_id: MultiInt,
    locale: Optional[str] = None,
    preferred_place_id: Optional[int] = None,
    all_names: Optional[bool] = None,
    **params,
) -> JsonResponse:
    """Get one or more taxa by ID

    .. rubric:: Notes

    * API reference: :v1:`GET /taxa/{id} <Taxa/get_taxa_id>`

    Example:
        >>> response = get_taxa_by_id(343248)
        >>> basic_fields = ['preferred_common_name', 'observations_count', 'wikipedia_url', 'wikipedia_summary']
        >>> print({f: response['results'][0][f] for f in basic_fields})
        {
            'preferred_common_name': 'Paper Wasps',
            'observations_count': 69728,
            'wikipedia_url': 'http://en.wikipedia.org/wiki/Polistinae',
            'wikipedia_summary': 'The Polistinae are eusocial wasps closely related to yellow jackets...',
        }

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_taxa_by_id.py

    Args:
        taxon_id: Get taxon with this ID. Multiple values are allowed.
        locale: Locale preference for taxon common names
        preferred_place_id: Place preference for regional taxon common names
        all_names: Include all taxon names in the response

    Returns:
        Response dict containing taxon records
    """
    response = get(
        f'{API_V1}/taxa',
        ids=taxon_id,
        locale=locale,
        preferred_place_id=preferred_place_id,
        all_names=all_names,
        **params,
    )
    taxa = response.json()
    taxa['results'] = convert_all_timestamps(taxa['results'])
    return taxa


@document_request_params(docs._taxon_params)
def get_taxa_autocomplete(**params) -> JsonResponse:
    """Given a query string, return taxa with names starting with the search term

    .. rubric:: Notes

    * API reference: :v1:`GET /taxa/autocomplete <Taxa/get_taxa_autocomplete>`
    * There appears to currently be a bug in the API that causes ``per_page`` to not have any effect.

    Example:
        Get just the name of the first matching taxon:

        >>> response = get_taxa_autocomplete(q='vespi')
        >>> print(response['results'][0]['name'])
        'Vespidae'

        Get basic info for taxa in response:

        >>> pprint(response)
        [52747   ]       Family: Vespidae (Hornets, Paper Wasps, Potter Wasps, and Allies)
        [84738   ]    Subfamily: Vespinae (Hornets and Yellowjackets)
        [131878  ]      Species: Nicrophorus vespillo (Vespillo Burying Beetle)

        If you get unexpected matches, the search likely matched a synonym, either in the form of a
        common name or an alternative classification. Check the ``matched_term`` property for more
        info. For example:

        >>> first_result = get_taxa_autocomplete(q='zygoca')['results'][0]
        >>> first_result["name"]
        "Schlumbergera truncata"  # This doesn't look like our search term!
        >>> first_result["matched_term"]
        "Zygocactus truncatus"    # ...Because it matched an older synonym for Schlumbergera

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_taxa_autocomplete.py

        .. dropdown:: Example Response (formatted)
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_taxa_autocomplete_minified.py

    Returns:
        Response dict containing taxon records
    """
    params = convert_rank_range(params)
    response = get(f'{API_V1}/taxa/autocomplete', **params)
    return response.json()


def get_taxa_map_layers(taxon_id: int, **params) -> JsonResponse:
    """Get some additional taxon metadata, including:
    * GBIF taxon ID and URL
    * Whether the taxon has range data and/or listed places

    Example:
        >>> response = get_taxa_map_layers(343248)

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_taxa_map_layers.json
                :language: JSON

    Args:
        taxon_id: iNaturalist taxon ID. Only one value is allowed.
    """
    response = get(f'{API_V1}/taxa/{taxon_id}/map_layers', **params).json()
    response['gbif_url'] = f'https://www.gbif.org/species/{response["gbif_id"]}'
    return response


def get_life_list_metadata(
    user_id: IntOrStr, locale: Optional[str] = None, **params
) -> JsonResponse:
    """Get common names and default taxon photos for a user's life list

    .. rubric:: Notes

    * Undocumented in API reference
    * On iNaturalist.org, this is used for dynamic life lists

    Args:
        user_id: iNaturalist user ID or username
        locale: Locale preference for taxon common names

    Example:
        >>> response = get_life_list_metadata(user_id='my_username')

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_lifelist_metadata.json
                :language: JSON

    Returns:
        Response dict containing taxon common names and photo URLs
    """
    return get(
        f'{API_V1}/taxa/lifelist_metadata', observed_by_user_id=user_id, locale=locale, **params
    ).json()



from pyinaturalist.constants import API_V1, JsonResponse
from pyinaturalist.converters import convert_all_coordinates, convert_all_timestamps
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.session import get


@document_request_params(docs._search_params, docs._pagination)
def search(q: str, **params) -> JsonResponse:
    """A unified search endpoint for places, projects, taxa, and/or users

    .. rubric:: Notes

    * API reference: :v1:`GET /search <Search/get_search>`

    Example:
        >>> response = search(q='odonat')
        >>> pprint(response)
        [Taxon  ] [47792   ] Order: Odonata (Dragonflies and Damselflies)
        [Place  ] [113562  ] Odonates of Peninsular India and Sri Lanka
        [Project] [9978    ] Ohio Dragonfly Survey  (Ohio Odonata Survey)
        [User   ] [113886  ] odonatanb (Gilles Belliveau)

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_search.py

    Returns:
        Response dict containing search results
    """
    response = get(f'{API_V1}/search', q=q, **params)
    search_results = response.json()
    search_results['results'] = convert_all_timestamps(search_results['results'])
    search_results['results'] = convert_all_coordinates(search_results['results'])
    return search_results



from logging import getLogger
from typing import Optional

from pyinaturalist.constants import API_V1, IntOrStr, JsonResponse
from pyinaturalist.converters import convert_all_timestamps, convert_generic_timestamps
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.session import get

logger = getLogger(__name__)


def get_user_by_id(user_id: IntOrStr, **params) -> JsonResponse:
    """Get a user by ID

    .. rubric:: Notes

    * API reference: :v1:`GET /users/{id} <Users/get_users_id>`

    Args:
        user_id: Get the user with this ID. Only a single ID is allowed per request.

    Example:
        >>> response = get_user_by_id(123456)
        >>> pprint(response)
        [1234] my_username

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_user_by_id.py

    Returns:
        Response dict containing user record
    """
    response = get(f'{API_V1}/users', ids=user_id, allow_str_ids=True, **params)
    results = response.json()['results']
    if not results:
        return {}
    return convert_generic_timestamps(results[0])


@document_request_params(docs._search_query, docs._project_id, docs._pagination)
def get_users_autocomplete(q: str, **params) -> JsonResponse:
    """Given a query string, return users with names or logins starting with the search term

    .. rubric:: Notes

    * API reference: :v1:`GET /users/autocomplete <Users/get_users_autocomplete>`
    * Pagination is supported; default page size is 6, and max is 100.

    Example:
        >>> response = get_users_autocomplete(q='my_userna')
        >>> pprint(response)
        [1234] my_username
        [12345] my_username_2

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_users_autocomplete.py

    Returns:
        Response dict containing user records
    """
    response = get(f'{API_V1}/users/autocomplete', q=q, **params)
    users = response.json()
    users['results'] = convert_all_timestamps(users['results'])
    return users


def get_current_user(access_token: Optional[str] = None, **params) -> JsonResponse:
    """Get your own user profile

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`GET /users/me <Users/get_users_me>`


    Example:
        >>> response = get_current_user(access_token=access_token)
        >>> pprint(response)
        [1234]  my_username

    Returns:
        Response dict containing a single user record
    """
    response = get(f'{API_V1}/users/me', access_token=access_token, **params)
    return convert_generic_timestamps(response.json()['results'][0])



from typing import Callable, List, Optional, Union

from pyinaturalist.constants import (
    API_V1,
    V1_OBS_ORDER_BY_PROPERTIES,
    HistogramResponse,
    IntOrStr,
    MultiFile,
    MultiInt,
    MultiIntOrStr,
)
from pyinaturalist.controllers import BaseController
from pyinaturalist.converters import ensure_list
from pyinaturalist.docs import copy_doc_signature
from pyinaturalist.docs import templates as docs
from pyinaturalist.models import (
    Annotation,
    ControlledTermCounts,
    LifeList,
    Observation,
    Photo,
    Sound,
    TaxonCounts,
    TaxonSummary,
    UserCounts,
)
from pyinaturalist.paginator import IDPaginator, IDRangePaginator, Paginator
from pyinaturalist.request_params import validate_multiple_choice_param
from pyinaturalist.v1 import (
    create_observation,
    delete_observation,
    get_life_list_metadata,
    get_observation_histogram,
    get_observation_identifiers,
    get_observation_observers,
    get_observation_popular_field_values,
    get_observation_species_counts,
    get_observation_taxon_summary,
    get_observation_taxonomy,
    update_observation,
    upload,
)

IDS_PER_REQUEST = 30


class ObservationController(BaseController):
    """:fa:`binoculars` Controller for Observation requests"""

    def __call__(self, observation_id: int, **params) -> Optional[Observation]:
        """Get a single observation by ID

        Example:
            >>> client.observations(16227955)

        Args:
            observation_ids: A single observation ID
        """
        return self.from_ids(observation_id, **params).one()

    def from_ids(self, observation_ids: MultiInt, **params) -> Paginator[Observation]:
        """Get one or more observations by ID

        .. rubric:: Notes

        * :fas:`lock-open` :ref:`Optional authentication <auth>` (For private/obscured coordinates)
        * API reference: :v1:`GET /observations/{id} <Observations/get_observations_id>`

        Examples:
            >>> obs = client.observations.from_ids(16227955).all()
            >>> obs = client.observations.from_ids([16227955, 16227956]).all()

        Args:
            observation_id: One or more observation IDs
        """

        def get_observations_by_id(_observation_ids: MultiInt, **params):
            return self.client.session.request(
                'GET', f'{API_V1}/observations', ids=_observation_ids, **params
            ).json()

        params = self.client.add_defaults(get_observations_by_id, params)

        return IDPaginator(
            get_observations_by_id,
            Observation,
            ids=ensure_list(observation_ids),
            ids_per_request=IDS_PER_REQUEST,
            **params,
        )

    @copy_doc_signature(*docs._get_observations, docs._only_id)
    def search(self, **params) -> Paginator[Observation]:
        """Search observations

        .. rubric:: Notes

        * :fas:`lock-open` :ref:`Optional authentication <auth>` (For private/obscured coordinates)
        * API reference: :v1:`GET /observations <Observations/get_observations>`

        Examples:

            Get observations of Monarch butterflies with photos + public location info,
            on a specific date in the province of Saskatchewan, CA (place ID 7953):

            >>> query = client.observations.search(
            >>>     taxon_name='Danaus plexippus',
            >>>     created_on='2020-08-27',
            >>>     photos=True,
            >>>     geo=True,
            >>>     geoprivacy='open',
            >>>     place_id=7953,
            >>> )
            >>> observations = query.all()

            Get basic info for observations in response:

            >>> pprint(observations)
            ID       Taxon ID Taxon                       Observed on   User      Location
            ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
            57707611 48662    Danaus plexippus (Monarch)  Aug 26, 2020  ingridt3  Michener Dr, Regina, SK, CA
            57754375 48662    Danaus plexippus (Monarch)  Aug 27, 2020  samroom   Railway Ave, Wilcox, SK, CA

            Search for observations with a given observation field:

            >>> obs = client.observations.search(observation_fields=['Species count']).all()

            Or observation field value:

            >>> obs = client.observations.search(observation_fields={'Species count': 2}).all()

        """

        def get_observations(**params):
            return self.client.session.get(f'{API_V1}/observations', **params).json()

        params = validate_multiple_choice_param(params, 'order_by', V1_OBS_ORDER_BY_PROPERTIES)
        params = self.client.add_defaults(get_observations, params)

        return ObservationPaginator(
            get_observations,
            Observation,
            loop=self.client.loop,
            annotation_callback=self.client.annotations.lookup,
            **params,
        )

    # TODO: Does this need a model with utility functions, or is {datetime: count} sufficient?
    @copy_doc_signature(*docs._get_observations, docs._observation_histogram)
    def histogram(self, **params) -> HistogramResponse:
        """Search observations and return histogram data for the given time interval

        .. rubric:: Notes

        * API reference: :v1:`GET /observations/histogram <Observations/get_observations_histogram>`
        * Search parameters are the same as :py:func:`~pyinaturalist.v1.observations.get_observations()`,
          with the addition of ``date_field`` and ``interval``.
        * ``date_field`` may be either 'observed' (default) or 'created'.
        * Observed date ranges can be filtered by parameters ``d1`` and ``d2``
        * Created date ranges can be filtered by parameters ``created_d1`` and ``created_d2``
        * ``interval`` may be one of: 'year', 'month', 'week', 'day', 'hour', 'month_of_year', or
          'week_of_year'; spaces are also allowed instead of underscores, e.g. 'month of year'.
        * The year, month, week, day, and hour interval options will set default values for ``d1`` and
          ``created_d1``, to limit the number of groups returned. You can override those values if you
          want data from a longer or shorter time span.
        * The 'hour' interval only works with ``date_field='created'``

        Example:
            Get observations per month during 2020 in Austria (place ID 8057)

            >>> client.observations.histogram(
            >>>     interval='month',
            >>>     d1='2020-01-01',
            >>>     d2='2020-12-31',
            >>>     place_id=8057,
            >>> )

            .. dropdown:: Example Response (observations per month of year)
                :color: primary
                :icon: code-square

                .. literalinclude:: ../sample_data/get_observation_histogram_month_of_year.py

            .. dropdown:: Example Response (observations per month)
                :color: primary
                :icon: code-square

                .. literalinclude:: ../sample_data/get_observation_histogram_month.py
                    :lines: 3-

            .. dropdown:: Example Response (observations per day)
                :color: primary
                :icon: code-square

                .. literalinclude:: ../sample_data/get_observation_histogram_day.py
                    :lines: 3-

        Returns:
            Dict of ``{time_key: observation_count}``. Keys are ints for 'month of year' and\
            'week of year' intervals, and :py:class:`~datetime.datetime` objects for all other intervals.
        """
        return self.client.request(get_observation_histogram, **params)

    # TODO: Example response UserCounts object?
    @copy_doc_signature(*docs._get_observations)
    def identifiers(self, **params) -> UserCounts:
        """Get identifiers of observations matching the search criteria and the count of
        observations they have identified. By default, results are sorted by ID count in descending.

        .. rubric:: Notes

        * API reference: :v1:`GET /observations/identifiers <Observations/get_observations_identifiers>`
        * This endpoint will only return up to 500 results.

        Example:
            >>> client.observations.identifiers(place_id=72645)
        """
        response = self.client.request(get_observation_identifiers, **params)
        return UserCounts.from_json(response)

    def life_list(
        self, user_id: Optional[IntOrStr] = None, locale: Optional[str] = None, **params
    ) -> LifeList:
        """Get taxa from a user's dynamic life list

        .. rubric:: Notes

        * Results are returned in a flat list, but are ordered as they would be in a taxonomic tree

        Args:
            user_id: iNaturalist user ID or username
            locale: Locale preference for taxon common names
        """
        response = self.client.request(get_observation_taxonomy, user_id=user_id, **params)

        # Add additional metadata to a life list response; requires a user ID
        if user_id:
            metadata = self.client.request(
                get_life_list_metadata, user_id=user_id, locale=locale, **params
            )
            meta_by_id = {item['id']: item for item in metadata['results']}
            for taxon in response['results']:
                taxon.update(meta_by_id.get(taxon['id'], {}))

        return LifeList.from_json(response)

    # TODO: Example response UserCounts object?
    @copy_doc_signature(*docs._get_observations)
    def observers(self, **params) -> UserCounts:
        """Get observers of observations matching the search criteria and the count of
        observations and distinct taxa of rank species they have observed.

        .. rubric:: Notes

        * API reference: :v1:`GET /observations/observers <Observations/get_observations_observers>`
        * Options for ``order_by`` are 'observation_count' (default) or 'species_count'
        * This endpoint will only return up to 500 results
        * See this issue for more details: https://github.com/inaturalist/iNaturalistAPI/issues/235

        Example:
            >>> client.observations.observers(place_id=72645, order_by='species_count')
        """
        response = self.client.request(get_observation_observers, **params)
        return UserCounts.from_json(response)

    @copy_doc_signature(*docs._get_observations)
    def popular_fields(self, **params) -> ControlledTermCounts:
        """Get controlled terms values and a monthly histogram of observations matching the search
        criteria.

        .. rubric:: Notes

        * API reference: :v1:`GET /observations/popular_field_values
          <Observations/get_observations_popular_field_values>`

        Example:
            >>> client.observations.popular_fields(species_name='Danaus plexippus', place_id=24)
        """
        response = self.client.request(get_observation_popular_field_values, **params)
        return ControlledTermCounts.from_json(response)

    @copy_doc_signature(*docs._get_observations)
    def species_count(self, **params) -> int:
        """Get a total count of species (or other 'leaf taxa') associated with observations matching the search
        criteria.

        .. rubric:: Notes

        * API reference: :v1:`GET /observations/species_counts <Observations/get_observations_species_counts>`
        * **Leaf taxa** are the leaves of the taxonomic tree, like species, subspecies, variety, or form
        * This method returns only the combined total number of leaf taxa in matched observations;
          :py:meth:`.species_counts` returns the full list of taxa and the number of observations for each

        Example:
            >>> client.observations.species_count(taxon_id=52775, place_id=6853)
        """
        response = self.client.request(get_observation_species_counts, count_only=True, **params)
        return response['total_results']

    @copy_doc_signature(*docs._get_observations)
    def species_counts(self, **params) -> TaxonCounts:
        """Get all species (or other 'leaf taxa') associated with observations matching the search
        criteria, and the count of observations they are associated with.

        .. rubric:: Notes

        * API reference: :v1:`GET /observations/species_counts <Observations/get_observations_species_counts>`
        * **Leaf taxa** are the leaves of the taxonomic tree, like species, subspecies, variety, or form

        Example:
            >>> client.observations.species_counts(user_login='my_username', quality_grade='research')
        """
        response = self.client.request(get_observation_species_counts, **params)
        return TaxonCounts.from_json(response)

    def taxon_summary(self, observation_id: int, **params) -> TaxonSummary:
        """Get information about an observation's taxon, within the context of the observation's location

        .. rubric:: Notes

        * API reference: :v1:`GET /observations/{id}/taxon_summary <Observations/get_observations_id_taxon_summary>`

        Args:
            observation_id: Observation ID to get taxon summary for

        Example:
            >>> client.observations.taxon_summary(7849808)
        """
        response = self.client.request(
            get_observation_taxon_summary, observation_id=observation_id, **params
        )
        return TaxonSummary.from_json(response)

    # TODO: create observations with Observation objects; requires model updates
    @copy_doc_signature(docs._create_observation)
    def create(self, **params) -> Observation:
        """Create or update an observation.

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * API reference: :v1:`POST /observations <Observations/post_observations>`

        Example:
            >>> client.observations.create(
            ...     species_guess='Pieris rapae',
            ...     photos='~/observation_photos/2020_09_01_14003156.jpg',
            ...     observation_fields={297: 1},  # 297 is the obs. field ID for 'Number of individuals'
            ... )
        """
        response = self.client.request(create_observation, auth=True, **params)
        return Observation.from_json(response)

    def delete(self, observation_ids: MultiInt, **params):
        """Delete one or more observations

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * API reference: :v1:`DELETE /observations/{id} <Observations/delete_observations_id>`


        Example:
            >>> delete_observation(17932425, token)

        Args:
            observation_ids: One or more observation IDs
        """
        for observation_id in ensure_list(observation_ids):
            self.client.request(
                delete_observation,
                auth=True,
                observation_id=observation_id,
                **params,
            )

    @copy_doc_signature(docs._observation_id, docs._create_observation)
    def update(self, **params) -> Observation:
        """Update a single observation

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * API reference: :v1:`PUT /observations <Observations/put_observations>`

        Example:
            >>> client.observations.update(
            >>>     17932425,
            >>>     description='updated description!',
            >>> )
        """
        params = self.client.add_defaults(update_observation, params, auth=True)
        response = update_observation(**params)
        return Observation.from_json(response)

    def upload(
        self,
        observation_id: int,
        photos: Optional[MultiFile] = None,
        sounds: Optional[MultiFile] = None,
        photo_ids: Optional[MultiIntOrStr] = None,
        **params,
    ) -> List[Union[Photo, Sound]]:
        """Upload one or more local photo and/or sound files, and add them to an existing observation.

        You may also attach a previously uploaded photo by photo ID, e.g. if your photo contains
        multiple organisms and you want to create a separate observation for each one.

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * API reference: :v1:`POST /observation_photos <Observation_Photos/post_observation_photos>`

        Example:

            >>> client.observations.upload(
            ...     1234,
            ...     photos=['~/observations/2020_09_01_140031.jpg', '~/observations/2020_09_01_140042.jpg'],
            ...     sounds='~/observations/2020_09_01_140031.mp3',
            ...     photo_ids=[1234, 5678],
            ... )

            .. dropdown:: Example Response
                :color: primary
                :icon: code-square

                .. literalinclude:: ../sample_data/upload_photos_and_sounds.json
                    :language: JSON

        Args:
            observation_id: the ID of the observation
            photos: One or more image files, file-like objects, file paths, or URLs
            sounds: One or more audio files, file-like objects, file paths, or URLs
            photo_ids: One or more IDs of previously uploaded photos to attach to the observation
            access_token: Access token for user authentication, as returned by :func:`get_access_token()`

        Returns:
            :py:class:`.Photo` or :py:class:`.Sound` objects for each uploaded file
        """
        responses = self.client.request(
            upload,
            auth=True,
            observation_id=observation_id,
            photos=photos,
            sounds=sounds,
            photo_ids=photo_ids,
            **params,
        )
        response_objs: List[Union[Photo, Sound]] = []
        for response in responses:
            if 'photo' in response:
                response_objs.append(Photo.from_json(response))
            elif 'sound' in response:
                response_objs.append(Sound.from_json(response))
        return response_objs


class ObservationPaginator(IDRangePaginator):
    """Paginate through observation results by a range of IDs instead of standard pagination
    parameters

    This also optionally fills in missing annotation information for each observation, using results
    from ``GET /controlled_terms``.
    """

    def __init__(
        self,
        *args,
        annotation_callback: Callable[[List[Annotation]], List[Annotation]],
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.annotation_callback = annotation_callback

    def next_page(self) -> List[Observation]:
        observations = super().next_page()
        # Use cached controlled_terms lookup to fill in missing annotation details
        for obs in observations:
            obs.annotations = self.annotation_callback(obs.annotations)
        return observations



from datetime import datetime
from itertools import chain
from typing import Any, Dict, List, Tuple

from pyinaturalist.constants import (
    ALL_LICENSES,
    GEOPRIVACY_LEVELS,
    INAT_BASE_URL,
    QUALITY_GRADES,
    Coordinates,
    DateTime,
    TableRow,
)
from pyinaturalist.docs import extend_init_signature
from pyinaturalist.models import (
    Annotation,
    BaseModel,
    BaseModelCollection,
    Comment,
    IconPhoto,
    Identification,
    LazyProperty,
    ObservationFieldValue,
    Photo,
    ProjectObservation,
    Sound,
    Taxon,
    User,
    coordinate_pair,
    datetime_field,
    datetime_now_field,
    define_model,
    define_model_collection,
    define_model_custom_init,
    field,
    upper,
)


@define_model
class Application(BaseModel):
    """:fa:`display` An iNaturalist mobile or third-party application"""

    icon: str = field(default=None, doc='Application icon URL')
    name: str = field(default=None, doc='Application name')
    url: str = field(default=None, doc='Application URL')

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'name', 'url']


@define_model
class Flag(BaseModel):
    """:fa:`flag` A flag on an observation"""

    comment = field(default=None, doc='Flag comment')
    created_at: datetime = datetime_field(doc='Date and time the flag was created')
    flag: str = field(default=None, doc='Flag type')
    flaggable_content: str = field(default=None)
    flaggable_id: int = field(default=None)
    flaggable_type: str = field(default=None)
    flaggable_user_id: int = field(default=None)
    resolved: bool = field(default=None)
    resolver_id: int = field(default=None)
    resolved_at: datetime = datetime_field(doc='Date and time the flag was resolved')
    updated_at: datetime = datetime_field(doc='Date and time the flag was last updated')
    user: property = LazyProperty(User.from_json, type=User, doc='User that added the flag')

    @property
    def username(self) -> str:
        return self.user.login

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'flag', 'resolved', 'username']


@define_model
class QualityMetric(BaseModel):
    """:fa:`list-check` An observation quality metric added by a user to an observation"""

    agree: bool = field(default=None, doc='Indicates if the user agrees with this metric')
    metric: str = field(default=None, doc='Quality metric name')
    user: property = LazyProperty(User.from_json, type=User, doc='User that added the metric')

    @property
    def username(self) -> str:
        return self.user.login

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Metric': self.metric,
            'Agree': self.agree,
            'User': self.user.login,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'metric', 'agree', 'username']


@define_model
class Vote(BaseModel):
    """:fa:`check` A vote on a data quality assessment metric"""

    created_at: datetime = datetime_field(doc='Date and time the vote was created')
    user: property = LazyProperty(User.from_json, type=User, doc='User that added the vote')
    vote_flag: bool = field(default=None)
    vote_scope: str = field(default=None)

    @property
    def username(self) -> str:
        return self.user.login

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Flag': self.vote_flag,
            'User': self.user.login,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'vote_flag', 'username']


# Uses the same schema as votes
@define_model
class Fave(Vote):
    """:fa:`star` An observation favorited by a user"""


@define_model_custom_init
class Observation(BaseModel):
    """:fa:`binoculars` An observation, based the schema of
    :v1:`GET /observations <Observations/get_observations>`
    """

    created_at: datetime = datetime_now_field(doc='Date and time the observation was created')
    captive: bool = field(
        default=None, doc='Indicates if the organism is non-wild (captive or cultivated)'
    )
    community_taxon_id: int = field(default=None, doc='The current community identification taxon')
    context_geoprivacy: str = field(default=None)
    context_taxon_geoprivacy: str = field(default=None)
    context_user_geoprivacy: str = field(default=None)
    description: str = field(default=None, doc='Observation description')
    geoprivacy: str = field(default=None, options=GEOPRIVACY_LEVELS, doc='Location privacy level')
    identifications_count: int = field(default=0, doc='Total number of identifications')
    identifications_most_agree: bool = field(
        default=None, doc='Indicates if most identifications agree with the community ID'
    )
    identifications_most_disagree: bool = field(
        default=None, doc='Indicates if most identifications disagree with the community ID'
    )
    identifications_some_agree: bool = field(
        default=None, doc='Indicates if some identifications agree with the community ID'
    )
    license_code: str = field(
        default=None, converter=upper, options=ALL_LICENSES, doc='Creative Commons license code'
    )
    location: Coordinates = coordinate_pair()
    mappable: bool = field(default=None, doc='Indicates if the observation can be shown on a map')
    num_identification_agreements: int = field(
        default=None, doc='Total identifications that agree with the community ID'
    )
    num_identification_disagreements: int = field(
        default=None, doc='Total identifications that disagree with the community ID'
    )
    oauth_application_id: str = field(
        default=None, doc='ID of the OAuth application used to create the observation, if any'
    )
    obscured: bool = field(
        default=None,
        doc='Indicates if coordinates are obscured (showing a broad approximate location on the map)',
    )
    observed_on: DateTime = datetime_field(doc='Date and time the organism was observed')
    outlinks: List[Dict] = field(
        factory=list, doc='Linked observation pages on other sites (e.g., GBIF)'
    )
    out_of_range: bool = field(
        default=None, doc='Indicates if the taxon is observed outside of its known range'
    )
    owners_identification_from_vision: bool = field(
        default=None, doc="Indicates if the owner's ID was selected from computer vision results"
    )
    place_guess: str = field(default=None, doc='Place name determined from observation coordinates')
    place_ids: List[int] = field(
        factory=list, doc='Place IDs associated with the observation coordinates'
    )
    positional_accuracy: int = field(
        default=None, doc='GPS accuracy in meters (real accuracy, if obscured)'
    )
    preferences: Dict[str, Any] = field(
        factory=dict,
        doc='Any user observation preferences (related to community IDs, coordinate access, etc.)',
    )
    private_location: Coordinates = coordinate_pair(
        doc=':fa:`lock` Private location in ``(latitude, longitude)`` decimal degrees'
    )
    private_place_ids: List[int] = field(
        factory=list, doc=':fa:`lock` Place IDs associated with the private observation coordinates'
    )
    private_place_guess: str = field(
        default=None, doc=':fa:`lock` Place name determined from private observation coordinates'
    )
    project_ids: List[int] = field(factory=list, doc='All associated project IDs')
    project_ids_with_curator_id: List[int] = field(
        factory=list, doc='Project IDs with a curator identification for this observation'
    )
    project_ids_without_curator_id: List[int] = field(
        factory=list, doc='Project IDs without curator identification for this observation'
    )
    public_positional_accuracy: int = field(
        default=None, doc='GPS accuracy in meters (not accurate if obscured)'
    )
    quality_grade: str = field(default=None, options=QUALITY_GRADES, doc='Quality grade')
    reviewed_by: List[int] = field(
        factory=list, doc='IDs of users who have reviewed the observation'
    )
    site_id: int = field(
        default=None, doc='Site ID for iNaturalist network members, or ``1`` for inaturalist.org'
    )
    species_guess: str = field(
        default=None, doc="Taxon name from observer's initial identification"
    )
    tags: List[str] = field(factory=list, doc='Arbitrary user tags added to the observation')
    taxon_geoprivacy: str = field(default=None)
    updated_at: DateTime = datetime_field(doc='Date and time the observation was last updated')
    uri: str = field(default=None, doc='Link to observation details page')
    uuid: str = field(
        default=None, doc='Universally unique ID; generally preferred over ``id`` where possible'
    )
    viewer_trusted_by_observer: bool = field(
        default=None, doc='Observer trusts the authenticated user with access to hidden coordinates'
    )

    # Lazy-loaded model objects
    annotations: property = LazyProperty(
        Annotation.from_json_list, type=List[Annotation], doc='Observation annotations'
    )
    application: property = LazyProperty(
        Application.from_json, type=Application, doc='Application that created the observation'
    )
    comments: property = LazyProperty(
        Comment.from_json_list, type=List[Comment], doc='Observation comments'
    )
    faves: property = LazyProperty(
        Fave.from_json_list, type=List[Fave], doc='Users who have favorited the observation'
    )
    flags: property = LazyProperty(Flag.from_json_list, type=List[Flag], doc='Observation flags')
    identifications: property = LazyProperty(
        Identification.from_json_list, type=List[Identification], doc='Observation identifications'
    )
    ofvs: property = LazyProperty(
        ObservationFieldValue.from_json_list,
        type=List[ObservationFieldValue],
        doc='Observation field values',
    )
    photos: property = LazyProperty(
        Photo.from_json_list, type=List[Photo], doc='Observation photos'
    )
    project_observations: property = LazyProperty(
        ProjectObservation.from_json_list,
        type=List[ProjectObservation],
        doc='Details on any projects that the observation has been added to',
    )
    quality_metrics: property = LazyProperty(
        QualityMetric.from_json_list,
        type=List[QualityMetric],
        doc='Data quality assessment metrics',
    )
    sounds: property = LazyProperty(
        Sound.from_json_list, type=List[Sound], doc='Observation sound files'
    )
    taxon: property = LazyProperty(Taxon.from_json, type=Taxon, doc='Observation taxon')
    user: property = LazyProperty(User.from_json, type=User, doc='Observer')
    votes: property = LazyProperty(
        Vote.from_json_list, type=List[Vote], doc='Votes on data quality assessment metrics'
    )

    # Additional attributes from API response that aren't needed; just left here for reference
    # cached_votes_total: int = field(default=None)
    # comments_count: int = field(default=None)
    # created_at_details: Dict = field(factory=dict)
    # created_time_zone: str = field(default=None)
    # faves_count: int = field(default=None)
    # geojson: Dict = field(factory=dict)
    # id_please: bool = field(default=None)
    # map_scale: int = field(default=None)
    # non_owner_ids: List = field(factory=list)
    # non_traditional_projects: property = LazyProperty(Project.from_json_list, type=List[Project])
    # observed_on_details: Dict = field(factory=dict)
    # observed_on_string: str = field(default=None)
    # observation_photos: List[Photo] = field(converter=Photo.from_dict_list, factory=list)
    # observed_time_zone: str = field(default=None)
    # private_geojson: Dict = field(factory=dict)
    # spam: bool = field(default=None)
    # time_observed_at: DateTime = datetime_attr
    # time_zone_offset: str = field(default=None)

    # Attributes that will only be used during init and then omitted
    temp_attrs = ['time_observed_at']

    def __init__(self, **kwargs):
        # Convert observation timestamps prior to __attrs_init__
        observed_on = kwargs.pop('time_observed_at', None)
        if not isinstance(kwargs['observed_on'], datetime) and observed_on:
            kwargs['observed_on'] = observed_on

        # Set default URL based on observation ID
        if not kwargs.get('uri'):
            kwargs['uri'] = f'{INAT_BASE_URL}/observations/{kwargs.get("id", "")}'

        # Set identifications_count if missing
        if kwargs.get('identifications') and not kwargs.get('identifications_count'):
            kwargs['identifications_count'] = len(kwargs['identifications'])

        # Add ancestor info to Observation.taxon based on identification data, if available
        if kwargs.get('taxon') and (idents := kwargs.get('identifications')):
            try:
                taxa_by_id = {
                    t['id']: t
                    for t in chain.from_iterable(
                        [i.get('taxon', {}).get('ancestors', []) for i in idents]
                    )
                }
                ancestors = [taxa_by_id.get(i) for i in kwargs['taxon'].get('ancestor_ids')]
                kwargs['taxon']['ancestors'] = list(filter(None, ancestors))
            except AttributeError:
                pass

        self.__attrs_init__(**kwargs)

    @property
    def cumulative_ids(self) -> Tuple[int, int]:
        """Calculate the cumulative community ID score (agreements/total), as shown on the observation UI

        Returns:
            ``(agreements, total)``
        """
        idents_count = 0
        idents_agree = 0
        ident_taxon_ids = self.ident_taxon_ids

        for ident in filter(lambda i: i.current, self.identifications):
            user_taxon_ids = ident.taxon.ancestor_ids + [ident.taxon.id]
            if self.community_taxon_id in user_taxon_ids:
                # Count towards total & agree:
                if ident.taxon.id in ident_taxon_ids:
                    idents_count += 1
                    idents_agree += 1
            else:
                # Maverick counts against:
                idents_count += 1

        return (idents_agree, idents_count)

    @property
    def default_photo(self) -> Photo:
        """Get the default observation photo if any; otherwise use taxon default photo or icon"""
        if self.photos:
            return self.photos[0]
        elif self.taxon:
            return self.taxon.icon
        else:
            return IconPhoto.from_iconic_taxon('unknown')

    @property
    def ident_taxon_ids(self) -> List[int]:
        """Get all taxon IDs (including ancestors) from identifications"""
        ident_taxa = [
            ident.taxon for ident in self.identifications if ident.taxon and ident.current
        ]
        ident_ids = chain.from_iterable([t.ancestor_ids + [t.id] for t in ident_taxa])
        return list(set(ident_ids))

    @property
    def username(self) -> str:
        return self.user.login

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Taxon ID': self.taxon.id if self.taxon else None,
            'Taxon': self.taxon.rich_full_name if self.taxon else None,
            'Observed on': self.observed_on,
            'User': self.user.login,
            'Location': self.place_guess or self.location,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'taxon', 'observed_on', 'username', 'place_guess']


@define_model_collection
class Observations(BaseModelCollection):
    """:fa:`binoculars` A collection of observations"""

    data: List[Observation] = field(factory=list, converter=Observation.from_json_list)

    @property
    def identifiers(self) -> List[User]:
        """Get all unique identifiers"""
        unique_users = {i.user.id: i.user for obs in self.data for i in obs.identifications}
        return list(unique_users.values())

    @property
    def observers(self) -> List[User]:
        """Get all unique observers"""
        unique_users = {obs.user.id: obs.user for obs in self.data}
        return list(unique_users.values())

    @property
    def photos(self) -> List[Photo]:
        """Get default photo for each observation"""
        return [obs.default_photo for obs in self.data]

    @property
    def taxa(self) -> List[Taxon]:
        """Get all unique taxa"""
        unique_taxa = {obs.taxon.id: obs.taxon for obs in self.data}
        return list(unique_taxa.values())


# Fix __init__ and class docstring
Observation = extend_init_signature(Observation.__attrs_init__)(Observation)  # type: ignore



# ruff: noqa: F401, F403
"""Utilities for generating pyinaturalist documentation, including:

* Dynamic docs: Function signatures + docstrings based on API request params
* Static docs: Sphinx documentation on readthedocs.io
"""

from pyinaturalist.docs.docstrings import ApiDocstring, copy_annotations, copy_docstrings
from pyinaturalist.docs.emoji import EMOJI
from pyinaturalist.docs.signatures import (
    copy_doc_signature,
    copy_signatures,
    document_common_args,
    document_controller_description,
    document_controller_params,
    document_request_params,
    extend_init_signature,
)



from typing import Any

from pyinaturalist.constants import API_V0, JsonResponse
from pyinaturalist.converters import convert_all_timestamps
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.session import get, put


@document_request_params(docs._search_query, docs._pagination)
def get_observation_fields(**params) -> JsonResponse:
    """Search observation fields. Observation fields are basically typed data fields that
    users can attach to observation.

    .. rubric:: Notes

    * API reference: :v0:`GET /observation_fields <get-observation_fields>`
    * This endpoint does not support the ``per_page`` parameter

    Example:

        >>> get_observation_fields(q='number of individuals')
        >>> # Show just observation field IDs and names
        >>> from pprint import pprint
        >>> pprint({r['id']: r['name'] for r in response})

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_fields.py

    Returns:
        Observation fields
    """
    all_pages = params.pop('page', None)
    params['page'] = 1
    params['per_page'] = 30  # This seems to return 30 results regardless of the per_page param
    obs_fields = get(f'{API_V0}/observation_fields.json', **params).json()

    # This endpoint returns a list with no pagination metadata, so iterate until there are no more pages
    if all_pages:
        page_results = obs_fields
        while len(page_results) == 30:
            params['page'] += 1
            page_results = get(f'{API_V0}/observation_fields.json', **params).json()
            obs_fields.extend(page_results)

    obs_fields = convert_all_timestamps(obs_fields)
    return {'results': obs_fields, 'total_results': len(obs_fields)}


@document_request_params(docs._ofvs, docs._access_token)
def put_observation_field_values(
    observation_id: int, observation_field_id: int, value: Any, **params
) -> JsonResponse:
    """Set an observation field value on an observation

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v0:`PUT /observation_field_values/{id} <put-observation_field_values-id>`
    * The request will fail if this observation field is already set for this observation
    * To find an ``observation_field_id``, either user :py:func:`.get_observation_fields` or
      `search observation fields on iNaturalist <https://www.inaturalist.org/observation_fields>`_

    Example:
        >>> # First find an observation field by name, if the ID is unknown
        >>> response = get_observation_fields('vespawatch_id')
        >>> observation_field_id = response[0]['id']

        >>> put_observation_field_values(
        ...     observation_id=7345179,
        ...     observation_field_id=observation_field_id,
        ...     value=250,
        ...     access_token=token,
        ... )

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/post_put_observation_field_value.json
                :language: javascript

    Returns:
        The newly updated field value record
    """
    body = {
        'observation_field_value': {
            'observation_id': observation_id,
            'observation_field_id': observation_field_id,
            'value': value,
        }
    }
    response = put(f'{API_V0}/observation_field_values/{observation_field_id}', json=body, **params)
    return response.json()



from datetime import datetime
from typing import List

from pyinaturalist.constants import ID_CATEGORIES, TableRow
from pyinaturalist.models import (
    BaseModel,
    LazyProperty,
    Taxon,
    User,
    datetime_now_field,
    define_model,
    field,
)


@define_model
class Comment(BaseModel):
    """:fa:`comment` An observation comment, based on the schema of comments
    from `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_.
    """

    body: str = field(default='', doc='Comment text')
    created_at: datetime = datetime_now_field(doc='Date and time the comment was created')
    hidden: bool = field(default=None, doc='Indicates if the comment is hidden')
    uuid: str = field(default=None, doc='Universally unique identifier')
    user: property = LazyProperty(
        User.from_json, type=User, doc='User that added the comment or ID'
    )

    # Unused attributes
    # created_at_details: Dict = field(factory=dict)
    # flags: List = field(factory=list)
    # moderator_actions: List = field(factory=list)

    @property
    def truncated_body(self):
        """Comment text, truncated"""
        truncated_body = self.body.replace('\n', ' ').strip()
        if len(truncated_body) > 50:
            truncated_body = truncated_body[:47].strip() + '...'
        return truncated_body

    @property
    def username(self) -> str:
        return self.user.login if self.user else ''

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'User': self.username,
            'Created at': self.created_at,
            'Comment': self.truncated_body,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'username', 'created_at', 'truncated_body']


@define_model
class Identification(Comment):
    """:fa:`fingerprint` An observation identification, based on the schema of
    `GET /identifications <https://api.inaturalist.org/v1/docs/#!/Identifications/get_identifications>`_.
    """

    category: str = field(default=None, options=ID_CATEGORIES, doc='Identification category')
    current: bool = field(
        default=None, doc='Indicates if the identification is the currently accepted one'
    )
    current_taxon: bool = field(default=None)
    disagreement: bool = field(
        default=None, doc='Indicates if this identification disagrees with previous ones'
    )
    own_observation: bool = field(
        default=None, doc='Indicates if the identifier is also the observer'
    )
    previous_observation_taxon_id: int = field(default=None, doc='Previous observation taxon ID')
    taxon_change: bool = field(default=None)
    vision: bool = field(
        default=None, doc='Indicates if the taxon was selected from computer vision results'
    )
    taxon: property = LazyProperty(Taxon.from_json, type=Taxon, doc='Identification taxon')

    @property
    def taxon_name(self) -> str:
        return self.taxon.full_name

    # Unused attributes
    # created_at_details: {}
    # spam: bool = field(default=None)
    # flags: List = field(factory=list)
    # moderator_actions: List = field(factory=list)
    # observation: {}
    # taxon_id: int = field(default=None)

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Taxon ID': self.taxon.id,
            'Taxon': f'{self.taxon.emoji} {self.taxon.rich_full_name}',
            'User': self.user.login,
            'Category': self.category.title(),
            'From CV': self.vision,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'username', 'taxon_name', 'created_at', 'truncated_body']



from functools import wraps
from typing import Optional

from pyinaturalist.constants import API_V1, JsonResponse, MultiInt
from pyinaturalist.converters import convert_all_coordinates, convert_all_place_coordinates
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.paginator import AutocompletePaginator, JsonPaginator
from pyinaturalist.session import get


def get_places_by_id(place_id: MultiInt, **params) -> JsonResponse:
    """Get one or more places by ID

    .. rubric:: Notes

    * API reference: :v1:`GET /places/{id} <Places/get_places_id>`

    Example:
        >>> response = get_places_by_id([67591, 89191])
        >>> pprint(response)
        [89191] Conservation Area Riversdale
        [67591] Riversdale Beach

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_places_by_id.py

    Args:
        place_id: Get a place with this ID. Multiple values are allowed.

    Returns:
        Response dict containing place records
    """
    response = get(f'{API_V1}/places', ids=place_id, **params)

    # Convert coordinates to floats
    places = response.json()
    places['results'] = convert_all_coordinates(places['results'])
    return places


@document_request_params(docs._bounding_box, docs._name)
def get_places_nearby(
    nelat: float, nelng: float, swlat: float, swlng: float, **params
) -> JsonResponse:
    """Search for places near a given location

    .. rubric:: Notes

    * API reference: :v1:`GET /places/nearby <get_places_nearby>`

    Example:
        >>> bounding_box = (150.0, -50.0, -149.999, -49.999)
        >>> response = get_places_nearby(*bounding_box)

        Response is split into standard (curated) places and community (non-curated) places:

        >>> print(len(response['results']['standard']))
        10
        >>> print(len(response['results']['community']))
        10

        Show basic info for all places in response:

        >>> pprint(response)
        [97394] North America
        [97395] Asia
        [97393] Oceania
        ...

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_places_nearby.py

    Returns:
        Response dict containing place records, divided into 'standard' and 'community' places.
    """
    response = get(
        f'{API_V1}/places/nearby', nelat=nelat, nelng=nelng, swlat=swlat, swlng=swlng, **params
    )
    return convert_all_place_coordinates(response.json())


@document_request_params(docs._search_query, docs._pagination)
def get_places_autocomplete(q: Optional[str] = None, **params) -> JsonResponse:
    """Given a query string, get places with names starting with the search term

    .. rubric:: Notes

    * API reference: :v1:`GET /places/autocomplete <Places/get_places_autocomplete>`
    * This endpoint accepts a ``per_page`` param, up to a max of 20 (default 10)
    * Pages beyond the first page cannot be retrieved. Use ``page=all`` to attempt to retrieve
      additional results. See :py:func:`.paginate_autocomplete` for more info.

    Example:
        >>> response = get_places_autocomplete('Irkutsk')
        >>> pprint(response)
        [11803] Irkutsk
        [41854] Irkutskiy rayon
        [166186] Irkutsk Oblast - ADD
        [163077] Irkutsk agglomeration

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_places_autocomplete.py

    Returns:
        Response dict containing place records
    """
    if params.get('page') == 'all':
        places = PlaceAutocompletePaginator(q=q, **params).all()
    else:
        places = get(f'{API_V1}/places/autocomplete', q=q, **params).json()

    places['results'] = convert_all_coordinates(places['results'])
    return places


class PlaceAutocompletePaginator(JsonPaginator, AutocompletePaginator):  # type: ignore
    def __init__(self, **kwargs):
        kwargs['per_page'] = 20

        @wraps(get)
        def request_function(**request_kwargs):
            return get(f'{API_V1}/places/autocomplete', **request_kwargs)

        super().__init__(request_function, **kwargs)



from typing import List, Union

from pyinaturalist.constants import TableRow
from pyinaturalist.models import BaseModel, Place, Project, Taxon, User, define_model, field

SEARCH_RESULT_TYPES = {cls.__name__: cls for cls in [Place, Project, Taxon, User]}
SEARCH_RESULT_TITLES = {'Place': 'name', 'Project': 'title', 'Taxon': 'full_name', 'User': 'login'}
SearchResultRecord = Union[Place, Project, Taxon, User]


@define_model
class SearchResult(BaseModel):
    """:fa:`search` A search result of any type, based on the schema of
    `GET /search <https://api.inaturalist.org/v1/docs/#!/Search/get_search>`_.
    """

    score: float = field(default=0, doc='Search result rank')
    type: str = field(default=None, options=SEARCH_RESULT_TYPES, doc='Search result type')
    matches: List[str] = field(factory=list, doc='Search terms matched')
    record: SearchResultRecord = field(default=None, doc='Search result object')

    # Convert value by datatype
    def __attrs_post_init__(self):
        if self.type in SEARCH_RESULT_TYPES and self.record is not None:
            result_cls = SEARCH_RESULT_TYPES[self.type]
            self.record = result_cls.from_json(self.record)

    @property
    def record_name(self) -> str:
        """Alias for type-specific name/title field"""
        name_attr = SEARCH_RESULT_TITLES[self.type]
        return getattr(self.record, name_attr)

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.record.id,
            'Type': self.type,
            'Score': f'{self.score:.2f}',
            'Name': self.record_name,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'type', 'score', 'record_name']



# TODO: pprint examples are out of date
from copy import deepcopy
from logging import getLogger
from typing import Optional

from pyinaturalist.constants import (
    API_V1,
    V1_OBS_ORDER_BY_PROPERTIES,
    HistogramResponse,
    JsonResponse,
    ListResponse,
    MultiFile,
    MultiInt,
    MultiIntOrStr,
    ResponseResult,
)
from pyinaturalist.converters import (
    convert_all_coordinates,
    convert_all_timestamps,
    convert_generic_timestamps,
    convert_histogram,
    convert_observation_histogram,
    convert_observation_timestamps,
    ensure_list,
)
from pyinaturalist.docs import document_common_args, document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.exceptions import ObservationNotFound
from pyinaturalist.paginator import paginate_all
from pyinaturalist.request_params import convert_observation_params, validate_multiple_choice_param
from pyinaturalist.session import delete, get, post, put

logger = getLogger(__name__)


@document_request_params(
    *docs._get_observations, docs._pagination, docs._only_id, docs._access_token
)
def get_observations(**params) -> JsonResponse:
    """Search observations

    .. rubric:: Notes

    * :fas:`lock-open` :ref:`Optional authentication <auth>` (For private/obscured coordinates)
    * API reference: :v1:`GET /observations <Observations/get_observations>`

    Examples:

        Get observations of Monarch butterflies with photos + public location info,
        on a specific date in the province of Saskatchewan, CA (place ID 7953):

        >>> response = get_observations(
        >>>     taxon_name='Danaus plexippus',
        >>>     created_on='2020-08-27',
        >>>     photos=True,
        >>>     geo=True,
        >>>     geoprivacy='open',
        >>>     place_id=7953,
        >>> )

        Get basic info for observations in response:

        >>> pprint(response)
        ID       Taxon ID Taxon                       Observed on   User      Location
        ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
        57707611 48662    Danaus plexippus (Monarch)  Aug 26, 2020  ingridt3  Michener Dr, Regina, SK, CA
        57754375 48662    Danaus plexippus (Monarch)  Aug 27, 2020  samroom   Railway Ave, Wilcox, SK, CA

        Search for observations with a given observation field:

        >>> response = get_observations(observation_fields=['Species count'])

        Or observation field value:

        >>> response = get_observations(observation_fields={'Species count': 2})

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observations_node.py

    Returns:
        Response dict containing observation records
    """
    params = validate_multiple_choice_param(params, 'order_by', V1_OBS_ORDER_BY_PROPERTIES)

    if params.get('page') == 'all':
        observations = paginate_all(get, f'{API_V1}/observations', method='id', **params)
    else:
        observations = get(f'{API_V1}/observations', **params).json()

    observations['results'] = convert_all_coordinates(observations['results'])
    observations['results'] = convert_all_timestamps(observations['results'])
    return observations


@document_common_args
def get_observations_by_id(
    observation_id: MultiInt, access_token: Optional[str] = None, **params
) -> JsonResponse:
    """Get one or more observations by ID

    .. rubric:: Notes

    * :fas:`lock-open` :ref:`Optional authentication <auth>` (For private/obscured coordinates)
    * API reference: :v1:`GET /observations/{id} <Observations/get_observations_id>`
    * This endpoint returns more complete annotation details compared to
      :py:func:`~pyinaturalist.v1.observations.get_observations`.
      See :py:class:`.Annotation` for details.

    Example:
        >>> response = get_observations_by_id(16227955)
        >>> response = get_observations_by_id([16227955, 16227956])

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observations_by_id.py

    Args:
        observation_id: Get an observation with this ID. Multiple IDs are allowed.
        access_token: An access token, as returned by :py:func:`.get_access_token()`

    Returns:
        Response dict containing observation records
    """
    observations = get(
        f'{API_V1}/observations', ids=observation_id, access_token=access_token, **params
    ).json()
    observations['results'] = convert_all_coordinates(observations['results'])
    observations['results'] = convert_all_timestamps(observations['results'])
    return observations


@document_request_params(*docs._get_observations, docs._observation_histogram)
def get_observation_histogram(**params) -> HistogramResponse:
    """Search observations and return histogram data for the given time interval

    .. rubric:: Notes

    * API reference: :v1:`GET /observations/histogram <Observations/get_observations_histogram>`
    * Search parameters are the same as :py:func:`~pyinaturalist.v1.observations.get_observations()`,
      with the addition of ``date_field`` and ``interval``.
    * ``date_field`` may be either 'observed' (default) or 'created'.
    * Observed date ranges can be filtered by parameters ``d1`` and ``d2``
    * Created date ranges can be filtered by parameters ``created_d1`` and ``created_d2``
    * ``interval`` may be one of: 'year', 'month', 'week', 'day', 'hour', 'month_of_year', or
      'week_of_year'; spaces are also allowed instead of underscores, e.g. 'month of year'.
    * The year, month, week, day, and hour interval options will set default values for ``d1`` and
      ``created_d1``, to limit the number of groups returned. You can override those values if you
      want data from a longer or shorter time span.
    * The 'hour' interval only works with ``date_field='created'``

    Example:
        Get observations per month during 2020 in Austria (place ID 8057)

        >>> response = get_observation_histogram(
        >>>     interval='month',
        >>>     d1='2020-01-01',
        >>>     d2='2020-12-31',
        >>>     place_id=8057,
        >>> )

        .. dropdown:: Example Response (observations per month of year)
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_histogram_month_of_year.py

        .. dropdown:: Example Response (observations per month)
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_histogram_month.py
                :lines: 3-

        .. dropdown:: Example Response (observations per day)
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_histogram_day.py
                :lines: 3-

    Returns:
        Dict of ``{time_key: observation_count}``. Keys are ints for 'month of year' and\
        'week of year' intervals, and :py:class:`~datetime.datetime` objects for all other intervals.
    """
    response = get(f'{API_V1}/observations/histogram', **params)
    return convert_observation_histogram(response.json())


@document_request_params(*docs._get_observations, docs._pagination)
def get_observation_identifiers(**params) -> JsonResponse:
    """Get identifiers of observations matching the search criteria and the count of
    observations they have identified. By default, results are sorted by ID count in descending.

    .. rubric:: Notes

    * API reference: :v1:`GET /observations/identifiers <Observations/get_observations_identifiers>`
    * This endpoint will only return up to 500 results.

    Example:
        >>> response = get_observation_identifiers(place_id=72645)
        >>> pprint(response)
        [409010  ] jdoe42 (Jane Doe)
        [691216  ] jbrown252 (James Brown)
        [3959037 ] tnsparkleberry

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_identifiers_ex_results.json
                :language: JSON

    Returns:
        Response dict of identifiers
    """
    params.setdefault('per_page', 500)
    response = get(f'{API_V1}/observations/identifiers', **params)
    return response.json()


@document_request_params(*docs._get_observations, docs._pagination)
def get_observation_observers(**params) -> JsonResponse:
    """Get observers of observations matching the search criteria and the count of
    observations and distinct taxa of rank species they have observed.

    .. rubric:: Notes

    * API reference: :v1:`GET /observations/observers <Observations/get_observations_observers>`
    * Options for ``order_by`` are 'observation_count' (default) or 'species_count'
    * This endpoint will only return up to 500 results
    * See this issue for more details: https://github.com/inaturalist/iNaturalistAPI/issues/235

    Example:
        >>> response = get_observation_observers(place_id=72645, order_by='species_count')
        >>> pprint(response)
        [1566366 ] fossa1211
        [674557  ] schurchin
        [5813    ] fluffberger (Fluff Berger)


        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_observers_ex_results.json
                :language: JSON

    Returns:
        Response dict of observers
    """
    params.setdefault('per_page', 500)
    response = get(f'{API_V1}/observations/observers', **params)
    return response.json()


@document_request_params(*docs._get_observations)
def get_observation_popular_field_values(**params) -> JsonResponse:
    """Get controlled terms values and a monthly histogram of observations matching the search
    criteria.

    .. rubric:: Notes

    * API reference: :v1:`GET /observations/popular_field_values <Observations/get_observations_popular_field_values>`

    Example:
        >>> response = get_observation_popular_field_values(
        ...     species_name='Danaus plexippus', place_id=24,
        ... )

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_popular_field_values.py

    Returns:
        Response dict. Each record contains a ``count``, a ``month_of_year`` histogram, a
            ``controlled_attribute``, and a ``controlled_value``.
    """
    response_json = get(f'{API_V1}/observations/popular_field_values', **params).json()
    for r in response_json['results']:
        r['month_of_year'] = convert_histogram(r['month_of_year'], interval='month_of_year')
    return response_json


@document_request_params(*docs._get_observations, docs._pagination)
def get_observation_species_counts(**params) -> JsonResponse:
    """Get all species (or other 'leaf taxa') associated with observations matching the search
    criteria, and the count of observations they are associated with.

    .. rubric:: Notes

    * API reference: :v1:`GET /observations/species_counts <Observations/get_observations_species_counts>`
    * **Leaf taxa** are the leaves of the taxonomic tree, e.g., species, subspecies, variety, etc.

    Example:
        >>> response = get_observation_species_counts(user_login='my_username', quality_grade='research')
        >>> pprint(response)
        [62060] Species: Palomena prasina (Green Shield Bug): 10
        [84804] Species: Graphosoma italicum (European Striped Shield Bug): 8
        [55727] Species: Cymbalaria muralis (Ivy-leaved toadflax): 3
        ...

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_species_counts.py

    Returns:
        Response dict containing taxon records with counts
    """
    if params.get('page') == 'all':
        return paginate_all(get, f'{API_V1}/observations/species_counts', **params)
    else:
        return get(f'{API_V1}/observations/species_counts', **params).json()


@document_request_params(*docs._get_observations)
def get_observation_taxonomy(**params) -> JsonResponse:
    """Get observation counts for all taxa in observations matching the search criteria.

    .. rubric:: Notes

    * Undocumented in API reference
    * On iNaturalist.org, this is mainly used for dynamic life lists
    * Results are returned in a flat list, but are ordered as they would be in a taxonomic tree

    Example:
        >>> response = get_observation_taxonomy(user_id='my_username')

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_taxonomy.json
                :language: JSON

    Returns:
        Response dict containing taxon records with counts
    """
    return get(f'{API_V1}/observations/taxonomy', **params).json()


@document_common_args
def get_observation_taxon_summary(observation_id: int, **params) -> JsonResponse:
    """Get information about an observation's taxon, within the context of the observation's location

    .. rubric:: Notes

    * API reference: :v1:`GET /observations/{id}/taxon_summary <Observations/get_observations_id_taxon_summary>`

    Args:
        observation_id: Observation ID to get taxon summary for

    Example:
        >>> response = get_observation_taxon_summary(7849808)

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observation_taxon_summary.py

    Returns:
        Response dict containing taxon summary, optionally with conservation status and listed taxon
    """
    results = get(f'{API_V1}/observations/{observation_id}/taxon_summary', **params).json()
    results['conservation_status'] = convert_generic_timestamps(results['conservation_status'])
    results['listed_taxon'] = convert_generic_timestamps(results['listed_taxon'])
    return results


@document_request_params(docs._access_token, docs._create_observation)
def create_observation(**params) -> JsonResponse:
    """Create or update an observation.

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`POST /observations <Observations/post_observations>`

    Example:
        >>> token = get_access_token()
        >>> # Create a new observation:
        >>> create_observation(
        ...     access_token=token,
        ...     species_guess='Pieris rapae',
        ...     photos='~/observation_photos/2020_09_01_14003156.jpg',
        ...     observation_fields={297: 1},  # 297 is the obs. field ID for 'Number of individuals'
        ... )
        >>>
        >>> # Update an existing observation:
        >>> create_observation(
        ...     access_token=token,
        ...     uuid='53411fc2-bdf0-434e-afce-4dac33970173',
        ...     description='Updated description!',
        ... )

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/create_observation_v1.json
                :language: JSON

    Returns:
        JSON response containing the newly created observation(s)
    """
    photos, sounds, photo_ids, params, kwargs = convert_observation_params(params)
    response = post(f'{API_V1}/observations', json={'observation': params}, **kwargs)
    response_json = response.json()
    observation_id = response_json['id']

    upload(observation_id, photos=photos, sounds=sounds, photo_ids=photo_ids, **kwargs)
    return response_json


@document_request_params(
    docs._observation_id,
    docs._access_token,
    docs._create_observation,
)
def update_observation(observation_id: int, **params) -> ListResponse:
    """Update a single observation

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`PUT /observations <Observations/put_observations>`

    Example:

        >>> token = get_access_token()
        >>> update_observation(
        >>>     17932425,
        >>>     access_token=token,
        >>>     description='updated description!',
        >>> )

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/update_observation_result.json
                :language: javascript

    Returns:
        JSON response containing the newly updated observation(s)
    """
    photos, sounds, photo_ids, params, kwargs = convert_observation_params(params)
    payload = {'observation': params}

    # If adding photos by ID, they must be appended to the list of existing photo IDs
    if photo_ids:
        logger.info(f'Adding {len(photo_ids)} existing photos')
        obs = get_observation(observation_id)
        combined_photo_ids = [p['id'] for p in obs['photos']]
        combined_photo_ids.extend(ensure_list(photo_ids))
        payload['local_photos'] = {str(observation_id): combined_photo_ids}
        kwargs.pop('ignore_photos', None)

    response = put(f'{API_V1}/observations/{observation_id}', json=payload, **kwargs)
    upload(observation_id, photos=photos, sounds=sounds, **kwargs)
    return response.json()


@document_common_args
def upload(
    observation_id: int,
    photos: Optional[MultiFile] = None,
    sounds: Optional[MultiFile] = None,
    photo_ids: Optional[MultiIntOrStr] = None,
    **params,
) -> ListResponse:
    """Upload one or more local photo and/or sound files, and add them to an existing observation.

    You may also attach a previously uploaded photo by photo ID, e.g. if your photo contains
    multiple organisms and you want to create a separate observation for each one.

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`POST /observation_photos <Observation_Photos/post_observation_photos>`

    Example:

        >>> token = get_access_token()
        >>> upload(
        ...     1234,
        ...     photos=['~/observations/2020_09_01_140031.jpg', '~/observations/2020_09_01_140042.jpg'],
        ...     sounds='~/observations/2020_09_01_140031.mp3',
        ...     photo_ids=[1234, 5678],
        ...     access_token=token,
        ... )

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/upload_photos_and_sounds.json
                :language: JSON

    Args:
        observation_id: the ID of the observation
        photos: One or more image files, file-like objects, file paths, or URLs
        sounds: One or more audio files, file-like objects, file paths, or URLs
        photo_ids: One or more IDs of previously uploaded photos to attach to the observation
        access_token: Access token for user authentication, as returned by :func:`get_access_token()`

    Returns:
        Information about the uploaded file(s)
    """
    params['raise_for_status'] = False
    session = params.pop('session', None)
    responses = []
    photos, sounds = ensure_list(photos), ensure_list(sounds)
    logger.info(f'Uploading {len(photos)} photos and {len(sounds)} sounds')

    # Upload photos
    photo_params = deepcopy(params)
    photo_params['observation_photo[observation_id]'] = observation_id
    for photo in photos:
        response = post(
            f'{API_V1}/observation_photos', files=photo, session=session, **photo_params
        )
        responses.append(response)

    # Upload sounds
    sound_params = deepcopy(params)
    sound_params['observation_sound[observation_id]'] = observation_id
    for sound in sounds:
        response = post(
            f'{API_V1}/observation_sounds', files=sound, session=session, **sound_params
        )
        responses.append(response)

    # Attach previously uploaded photos by ID
    if photo_ids:
        response = update_observation(
            observation_id,
            photo_ids=photo_ids,
            session=session,
            access_token=params.get('access_token', None),
        )
        responses.append(response)

    # Wait until all uploads complete to raise errors for any failed uploads
    for response in responses:
        response.raise_for_status()
    return [response.json() for response in responses]


@document_request_params(docs._observation_id, docs._access_token)
def delete_observation(observation_id: int, access_token: Optional[str] = None, **params):
    """Delete an observation

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`DELETE /observations/{id} <Observations/delete_observations_id>`

    Example:
        >>> token = get_access_token()
        >>> delete_observation(17932425, token)

    Returns:
        If successful, no response is returned from this endpoint

    Raises:
        :py:exc:`.ObservationNotFound` if the requested observation doesn't exist
        :py:exc:`requests.HTTPError` (403) if the observation belongs to another user
    """
    response = delete(
        f'{API_V1}/observations/{observation_id}',
        access_token=access_token,
        raise_for_status=False,
        **params,
    )
    if response.status_code == 404:
        raise ObservationNotFound(response=response)
    response.raise_for_status()


@document_common_args
def get_observation(
    observation_id: int, access_token: Optional[str] = None, **params
) -> ResponseResult:
    """Get details about a single observation by ID

    .. rubric:: Notes

    * :fas:`triangle-exclamation` Deprecated; use :func:`get_observations`
      or :func:`get_observations_by_id` instead
    * :fas:`lock-opens` :ref:`Optional authentication <auth>` (For private/obscured coordinates)
    * API reference: :v1:`GET /observations/{id} <Observations/get_observations_id>`

    Example:
        >>> response = get_observation(16227955)

    Args:
        observation_id: Get the observation with this ID. Only a single value is allowed.
        access_token: An access token, as returned by :py:func:`.get_access_token()`

    Returns:
        A dict with details on the observation

    Raises:
        :py:exc:`.ObservationNotFound` If an invalid observation is specified
    """

    response = get_observations(id=observation_id, access_token=access_token, **params)
    if response['results']:
        return convert_observation_timestamps(response['results'][0])
    raise ObservationNotFound(response=response)



from typing import List, Optional

from pyinaturalist.constants import MultiInt, MultiStr
from pyinaturalist.controllers import BaseController
from pyinaturalist.models import SearchResult
from pyinaturalist.v1 import search


class SearchController(BaseController):
    """:fa:`search` Unified text search"""

    def __call__(
        self,
        q: str,
        sources: Optional[MultiStr] = None,
        place_id: Optional[MultiInt] = None,
        locale: Optional[str] = None,
        preferred_place_id: Optional[int] = None,
        **params,
    ) -> List[SearchResult]:
        """A unified text search endpoint for places, projects, taxa, and/or users

        .. rubric:: Notes

        * API reference: :v1:`GET /search <Search/get_search>`

        Example:
            >>> response = client.search(q='odonat')
            >>> pprint(response)
            ID        Type      Score   Name
            ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━
            47792     Taxon     9.45    Order Odonata (Dragonflies And Damselflies)
            113562    Place     7.70    Odonates of Peninsular India and Sri Lanka
            9978      Project   7.27    Ohio Dragonfly Survey (Ohio Odonata Survey)
            5665218   User      6.10    odonatachr

        Args:
            q: Search query
            sources: Object types to search
            place_id: Results must be associated with this place
            locale: Locale preference for taxon common names
            preferred_place_id: Place preference for regional taxon common names

        Returns:
            Response dict containing search results
        """
        response = search(
            q,
            sources=sources,
            place_id=place_id,
            locale=locale,
            preferred_place_id=preferred_place_id,
            **params,
        )
        return SearchResult.from_json_list(response)



"""Session class and related functions for preparing and sending API requests"""

import json
import threading
from collections import defaultdict
from importlib.metadata import version as pkg_version
from json import JSONDecodeError
from logging import getLogger
from os import getenv
from typing import Dict, Optional, Type

import pyrate_limiter
from requests import PreparedRequest, Request, Response, Session
from requests.adapters import HTTPAdapter
from requests.utils import default_user_agent
from requests_cache import (
    AnyRequest,
    CachedResponse,
    CacheMixin,
    ExpirationPatterns,
    ExpirationTime,
)
from requests_ratelimiter import (
    AbstractBucket,
    BucketFullException,
    Limiter,
    LimiterMixin,
    RequestRate,
    SQLiteBucket,
)
from urllib3.util import Retry

from pyinaturalist.constants import (
    CACHE_EXPIRATION,
    CACHE_FILE,
    CONNECT_TIMEOUT,
    DEFAULT_LOCK_PATH,
    MAX_DELAY,
    RATELIMIT_FILE,
    REQUEST_BURST_RATE,
    REQUEST_RETRIES,
    REQUEST_TIMEOUT,
    REQUESTS_PER_DAY,
    REQUESTS_PER_MINUTE,
    REQUESTS_PER_SECOND,
    RETRY_BACKOFF,
    RETRY_STATUSES,
    WRITE_HTTP_METHODS,
    FileOrPath,
    MultiInt,
    RequestParams,
)
from pyinaturalist.converters import ensure_file_obj
from pyinaturalist.formatters import format_request, format_response
from pyinaturalist.request_params import (
    convert_url_ids,
    preprocess_request_body,
    preprocess_request_params,
)

logger = getLogger('pyinaturalist')
thread_local = threading.local()


class ClientSession(CacheMixin, LimiterMixin, Session):
    """Custom session class used for sending API requests. Combines the following features and
    settings:

    * Caching
    * Rate-limiting (skipped for cached requests)
    * Retries
    * Timeouts
    """

    def __init__(
        self,
        cache_file: FileOrPath = CACHE_FILE,
        cache_control: bool = True,
        expire_after: Optional[ExpirationTime] = None,
        urls_expire_after: Optional[ExpirationPatterns] = None,
        per_second: float = REQUESTS_PER_SECOND,
        per_minute: float = REQUESTS_PER_MINUTE,
        per_day: float = REQUESTS_PER_DAY,
        burst: int = REQUEST_BURST_RATE,
        bucket_class: Type[AbstractBucket] = SQLiteBucket,
        backoff_factor: float = RETRY_BACKOFF,
        ratelimit_path: Optional[str] = RATELIMIT_FILE,
        lock_path: Optional[str] = DEFAULT_LOCK_PATH,
        max_retries: int = REQUEST_RETRIES,
        timeout: int = REQUEST_TIMEOUT,
        user_agent: Optional[str] = None,
        **kwargs,
    ):
        """Get a Session object, optionally with custom settings for caching and rate-limiting.

        Args:
            cache_file: Cache file path to use; defaults to the system default cache directory
            cache_control: Use server-provided Cache-Control headers to set cache expiration when
                possible (instead of ``expire_after`` or ``urls_expire_after``)
            expire_after: How long to keep cached API requests; for advanced options, see
                `requests-cache: Expiration <https://requests-cache.readthedocs.io/en/stable/user_guide/expiration.html>`_
            urls_expire_after Glob patterns for per-URL cache expiration; See
                `requests-cache: URL Patterns <https://requests-cache.readthedocs.io/en/stable/user_guide/expiration.html#url-patterns>`_
            per_second: Max requests per second
            per_minute: Max requests per minute
            per_day: Max requests per day
            burst: Max number of consecutive requests allowed before applying per-second rate-limiting
            bucket_class: Rate-limiting backend to use; defaults to a persistent SQLite database.
            backoff_factor: Factor for increasing delays between retries
            ratelimit_path: Path to SQLite database for rate-limiting;
                defaults to the system default cache directory
            lock_path: Path to file lock for multiprocess rate-limit database;
                 defaults to the system default cache directory
            max_retries: Maximum number of times to retry a failed request
            timeout: Maximum number of seconds to wait for a response from the server
            user_agent: Additional User-Agent info to pass to API requests
            kwargs: Additional keyword arguments for :py:class:`~requests_cache.session.CachedSession`
                and/or :py:class:`~requests_ratelimiter.requests_ratelimiter.LimiterSession`
        """
        # Extend default URL expiration patterns with user-provided ones, if any
        url_patterns = CACHE_EXPIRATION
        if urls_expire_after:
            url_patterns.update(urls_expire_after)
        self.timeout = timeout

        # Extra args to pass to rate limiter backend
        bucket_kwargs = kwargs.pop('bucket_kwargs', {})
        if ratelimit_path:
            bucket_kwargs['path'] = ratelimit_path
        if lock_path and bucket_class is FileLockSQLiteBucket:
            bucket_kwargs['lock_path'] = lock_path

        super().__init__(  # type: ignore  # false positive
            # Cache settings
            cache_name=cache_file,
            backend='sqlite',
            cache_control=cache_control,
            expire_after=expire_after,
            urls_expire_after=url_patterns,
            ignored_parameters=['Authorization', 'access_token'],
            stale_if_error=True,
            # Rate limit settings
            bucket_class=bucket_class,
            bucket_kwargs=bucket_kwargs,
            per_second=per_second,
            per_minute=per_minute,
            per_day=per_day,
            per_host=True,
            burst=burst,
            max_delay=MAX_DELAY,
            **kwargs,
        )

        # Separate rate limiter specific to forced refresh requests
        self.refresh_limiter = Limiter(
            RequestRate(1, 122),
            bucket_class=bucket_class,
            bucket_kwargs=bucket_kwargs,
        )

        # Retry settings
        self.retries = Retry(
            total=max_retries, backoff_factor=backoff_factor, status_forcelist=RETRY_STATUSES
        )
        adapter = HTTPAdapter(max_retries=self.retries)
        self.mount('https://', adapter)

        # Default headers
        self.headers['Accept'] = 'application/json'
        user_agent_details = [
            default_user_agent(),
            f'pyinaturalist/{pkg_version("pyinaturalist")}',
            user_agent or '',
        ]
        self.headers['User-Agent'] = ' '.join(user_agent_details).strip()

    def prepare_inat_request(
        self,
        method: str,
        url: str,
        access_token: Optional[str] = None,
        files: Optional[FileOrPath] = None,
        headers: Optional[Dict] = None,
        ids: Optional[MultiInt] = None,
        json: Optional[Dict] = None,
        params: Optional[RequestParams] = None,
        allow_str_ids: bool = False,
        **kwargs,
    ) -> Request:
        """Translate pyinaturalist-specific options into standard request arguments"""
        # Prepare request params and URL
        params = preprocess_request_params(params)
        url = convert_url_ids(url, ids, allow_str_ids)

        # Set auth header
        headers = headers or {}
        if access_token:
            headers['Authorization'] = f'Bearer {access_token}'

        # Convert any datetimes in request body, and read any files or URLs for uploading
        json = preprocess_request_body(json)
        if files:
            files = {'file': ensure_file_obj(files, self)}  # type: ignore

        # Convert into a PreparedRequest
        return Request(
            method=method,
            url=url,
            files=files,
            headers=headers,
            json=json,
            params=params,
            **kwargs,
        )

    def request(  # type: ignore
        self,
        method: str,
        url: str,
        headers: Optional[Dict] = None,
        json: Optional[Dict] = None,
        access_token: Optional[str] = None,
        allow_redirects: bool = False,
        allow_str_ids: bool = False,
        dry_run: bool = False,
        expire_after: Optional[ExpirationTime] = None,
        files: Optional[FileOrPath] = None,
        ids: Optional[MultiInt] = None,
        only_if_cached: bool = False,
        raise_for_status: bool = True,
        refresh: bool = False,
        stream: bool = False,
        timeout: Optional[int] = None,
        verify: bool = True,
        **params: RequestParams,
    ) -> Response:
        """Wrapper around :py:func:`requests.request` with additional options specific to iNat API requests

        Args:
            method: HTTP method
            url: Request URL
            headers: Request headers
            json: JSON request body
            access_token: access_token: the access token, as returned by :func:`get_access_token()`
            dry_run: Just log the request instead of sending a real request
            expire_after: How long to keep cached API requests
            files: File object, path, or URL to upload
            ids: One or more integer IDs used as REST resource(s) to request
            only_if_cached: Only return a response if it is cached
            raise_for_status: Raise an exception if the response status is not 2xx
            refresh: Skip reading from the cache and always fetch a fresh response
            stream: Stream the response content
            timeout: Time (in seconds) to wait for a response from the server; if exceeded, a
                :py:exc:`requests.exceptions.Timeout` will be raised.
            verify: Verify SSL certificates
            params: All other keyword arguments will be interpreted as request parameters

        Returns:
            API response
        """
        request = self.prepare_inat_request(
            method=method,
            url=url,
            access_token=access_token,
            allow_str_ids=allow_str_ids,
            files=files,
            headers=headers,
            ids=ids,
            json=json,
            params=params,
        )

        response = self.send(
            request,
            dry_run=dry_run,
            expire_after=expire_after,
            only_if_cached=only_if_cached,
            refresh=refresh,
            timeout=timeout,
            allow_redirects=allow_redirects,
            stream=stream,
            verify=verify,
        )

        # Raise an exception if the request failed (after retries are exceeded)
        if raise_for_status:
            response.raise_for_status()
        return response

    def send(  # type: ignore  # Adds kwargs not present in Session.send()
        self,
        request: AnyRequest,
        dry_run: bool = False,
        expire_after: Optional[ExpirationTime] = None,
        refresh: bool = False,
        retries: Optional[Retry] = None,
        timeout: Optional[int] = None,
        **kwargs,
    ) -> Response:
        """Send a request with caching, rate-limiting, and retries

        Args:
            request: Prepared request to send
            dry_run: Just log the request instead of sending a real request
            expire_after: How long to keep cached API requests
            refresh: Skip reading from the cache and always fetch a fresh response
            timeout: Maximum number of seconds to wait for a response from the server

        **Note:** :py:meth:`requests.Session.send` accepts separate timeout values for connect and
        read timeouts. The ``timeout`` argument will be used as the read timeout.
        """
        if not isinstance(request, PreparedRequest):
            request = super().prepare_request(request)
        logger.info(format_request(request, dry_run))

        # Make a mock request, if specified
        if dry_run or is_dry_run_enabled(request.method):
            return MockResponse(request)

        # Otherwise, send the request
        read_timeout = timeout or self.timeout
        response = super().send(
            request,
            expire_after=expire_after,
            refresh=refresh,
            timeout=(CONNECT_TIMEOUT, read_timeout),
            **kwargs,
        )
        response = self._validate_json(
            request,
            response,
            expire_after=expire_after,
            retries=retries,
            timeout=timeout,
            **kwargs,
        )

        logger.debug(format_response(response))
        return response

    def get_refresh_params(self, endpoint) -> Dict:
        """In some cases, we need to be sure we have the most recent version of a resource, for example
        when updating projects. Normally we would handle this with cache headers, but the CDN cache does
        not respect these if the cached response is less than ~2 minutes old.

        The iNat webapp handles this by adding an extra `v` request parameter, which results in a
        different cache key. This function does the same by tracking a separate rate limit per value,
        and finding the lowest value that is certain to result in a fresh response.
        """
        v = 0
        while True:
            try:
                bucket = f'{endpoint}?v={v}'
                self.refresh_limiter.try_acquire(bucket)
                break
            except BucketFullException as e:
                seconds = int(e.meta_info['remaining_time'])
                logger.debug(f'{bucket} cannot be refreshed again for {seconds} seconds')
                v += 1

        return {'refresh': True, 'v': v} if v > 0 else {'refresh': True}

    def _validate_json(
        self,
        request: PreparedRequest,
        response: Response,
        retries: Optional[Retry] = None,
        **kwargs,
    ) -> Response:
        """Occasionally, the API may return invalid (truncated) JSON, requiring a retry. This method
        checks for this condition, treats it as a request error, and applies existing retry settings
        (so behavior is consistent with ``urllib3.ConnectionPool.urlopen()``).
        """
        # Skip for non-JSON responses
        if not response.headers.get('Content-Type', '').startswith('application/json'):
            return response

        # Attempt to decode the response content as JSON
        try:
            response_json = response.json()
        # Update retry state and wait before sending the request again
        except JSONDecodeError as e:
            logger.info('Invalid JSON response; retrying...')
            retries = retries or self.retries
            retries = retries.increment(
                response.request.method,
                response.request.url,
                error=e,
            )
            retries.sleep()
            kwargs['force_refresh'] = True
            kwargs['retries'] = retries
            return self.send(request, **kwargs)
        # Save decoded JSON on response object, to avoid decoding twice
        else:
            response.json = lambda **kwargs: response_json  # type: ignore
            return response


class FileLockSQLiteBucket(pyrate_limiter.FileLockSQLiteBucket):
    """Bucket backed by a SQLite database and file lock. Suitable for usage from multiple processes
    with no shared state. Requires installing [py-filelock](https://py-filelock.readthedocs.io).

    The file lock is reentrant and shared across buckets, allowing a process to access multiple
    buckets at once.

    This modified class allows setting a custom lock path.
    """

    def __init__(self, lock_path: str = DEFAULT_LOCK_PATH, **kwargs):
        from filelock import FileLock

        super().__init__(**kwargs)
        self._lock = FileLock(lock_path)


def delete(url: str, session: Optional[ClientSession] = None, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.delete` with additional options specific to iNat API requests"""
    session = session or get_local_session()
    return session.request('DELETE', url, **kwargs)


def get(url: str, session: Optional[ClientSession] = None, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.get` with additional options specific to iNat API requests"""
    session = session or get_local_session()
    return session.request('GET', url, **kwargs)


def post(url: str, session: Optional[ClientSession] = None, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.post` with additional options specific to iNat API requests"""
    session = session or get_local_session()
    return session.request('POST', url, **kwargs)


def put(url: str, session: Optional[ClientSession] = None, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.put` with additional options specific to iNat API requests"""
    session = session or get_local_session()
    return session.request('PUT', url, **kwargs)


def clear_cache():
    """Clear all cached API responses"""
    get_local_session().cache.clear()


def env_to_bool(environment_variable: str) -> bool:
    """Translate an environment variable to a boolean value, accounting for minor
    variations (case, None vs. False, etc.)
    """
    env_value = getenv(environment_variable)
    return bool(env_value) and str(env_value).lower() not in ['false', 'none']


def get_local_session(**kwargs) -> ClientSession:
    """Get a thread-local Session object with default settings. This will be reused across requests
    to take advantage of connection pooling and (optionally) caching. If used in a multi-threaded
    context (for example, a :py:class:`~concurrent.futures.ThreadPoolExecutor`), this will create
    and store a separate session object for each thread.

    Args:
        kwargs: Keyword arguments for :py:func:`.ClientSession`
    """
    if not hasattr(thread_local, 'session'):
        thread_local.session = ClientSession(**kwargs)
    return thread_local.session


class MockResponse(CachedResponse):
    """A mock response to return in dry-run mode.
    This behaves the same as a cached response, but with the following additions:

    * Adds default response values
    * Returns a ``defaultdict`` when calling ``json()``, to accommodate checking for arbitrary keys
    """

    def __init__(self, request: Optional[PreparedRequest] = None, **kwargs):
        json_content = {
            'results': [],
            'total_results': 0,
            'access_token': '',
            'id': 'placeholder-id',
        }
        default_kwargs = {
            'headers': {'Cache-Control': 'no-store'},
            'request': request or PreparedRequest(),
            'status_code': 200,
            'reason': 'DRY_RUN',
            'content': json.dumps(json_content).encode(),
        }
        default_kwargs.update(kwargs)
        super().__init__(**default_kwargs)

    def json(self, **kwargs):
        return defaultdict(str, super().json(**kwargs))


def is_dry_run_enabled(method: str) -> bool:
    """A wrapper to determine if dry-run (aka test mode) has been enabled via either
    a constant or an environment variable. Dry-run mode may be enabled for either write
    requests, or all requests.
    """
    dry_run_enabled = env_to_bool('DRY_RUN_ENABLED')
    dry_run_write_only = env_to_bool('DRY_RUN_WRITE_ONLY') and method in WRITE_HTTP_METHODS
    return dry_run_enabled or dry_run_write_only



from logging import getLogger

from pyinaturalist.constants import API_V1, JsonResponse, MultiInt
from pyinaturalist.converters import convert_all_timestamps
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.session import get

logger = getLogger(__name__)


@document_request_params(docs._message_id, docs._access_token)
def get_message_by_id(message_id: MultiInt, **params) -> JsonResponse:
    """Get a message by ID

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`GET /messages/{id} <Messages/get_messages_id>`

    Example:
        >>> response = get_messages(123456)
        >>> pprint(response)

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_messages.json

    Returns:
        Response dict containing user record
    """
    response = get(f'{API_V1}/messages', ids=message_id, **params)
    messages = response.json()
    messages['results'] = convert_all_timestamps(messages['results'])
    return messages


@document_request_params(docs._message_params, docs._access_token)
def get_messages(**params) -> JsonResponse:
    """Get messages from the user's inbox

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`GET /messages <Messages/get_messages>`

    Example:
        >>> response = get_messages()
        >>> pprint(response)

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_messages.json

    Returns:
        Response dict containing user record
    """
    # `threads` is not compatible with `q` param, and includes totals from both inbox and sent
    if params.get('threads') is True:
        params['box'] = 'any'
        params['q'] = None

    response = get(f'{API_V1}/messages', **params)
    messages = response.json()
    messages['results'] = convert_all_timestamps(messages['results'])
    return messages


def get_unread_message_count(**params) -> int:
    """Get the number of unread messages in the user's inbox

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`GET /messages/unread <Messages/get_messages_unread>`

    Example:
        >>> get_unread_message_count()
        12

    Returns:
        Unread message count
    """
    response = get(f'{API_V1}/messages/unread', **params)
    try:
        return int(response.json()['count'])
    except (KeyError, TypeError, ValueError):
        logger.error(f'Failed to get unread message count: {response.text}', exc_info=True)
        return 0



from typing import Optional

from pyinaturalist.constants import API_V1, JsonResponse
from pyinaturalist.exceptions import TaxonNotFound
from pyinaturalist.session import get


def get_controlled_terms(taxon_id: Optional[int] = None, **params) -> JsonResponse:
    """List controlled terms and their possible values

    .. rubric:: Notes

    * API reference: :v1:`GET /controlled_terms <Controlled_Terms/get_controlled_terms>`

    Example:
        >>> response = get_controlled_terms()
        >>> pprint(response)
        1: Life Stage
            2: Adult
            3: Teneral
            4: Pupa
        ...

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_controlled_terms.json
                :language: JSON
    Returns:
        A dict containing details on controlled terms and their values
    """
    # For backwards-compatibility
    if taxon_id:
        return get_controlled_terms_for_taxon(taxon_id, **params)

    response = get(f'{API_V1}/controlled_terms', **params)
    return response.json()


def get_controlled_terms_for_taxon(taxon_id: int, **params) -> JsonResponse:
    """List controlled terms that are valid for the specified taxon.

    .. rubric:: Notes

    * API reference: :v1:`GET /controlled_terms/for_taxon <Controlled_Terms/get_controlled_terms_for_taxon>`

    Example:
        >>> response = get_controlled_terms_for_taxon(12345)
        >>> pprint(response)
        1: Life Stage
            2: Adult
            3: Teneral
            4: Pupa
        ...

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_controlled_terms_for_taxon.json
                :language: JSON
    Args:
        taxon_id: Taxon ID to get controlled terms for

    Returns:
        A dict containing details on controlled terms and their values

    Raises:
        :py:exc:`.TaxonNotFound`: If an invalid ``taxon_id`` is specified
    """
    response = get(f'{API_V1}/controlled_terms/for_taxon', taxon_id=taxon_id, **params)

    # This endpoint returns a 422 if the specified taxon does not exist
    if response.status_code in (404, 422):
        raise TaxonNotFound(response=response)
    return response.json()
